# -*- coding: utf-8 -*-
'''
Created on:    Nov 10, 2013
@author:        vahid
'''
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from isass.helpers import get_source_dirs, get_source_files, distinct, split_paths, minify
from isass import SassCompiler
import os.path
from isass.manifest import Manifest
import sys
import traceback


class IsassEventHandler(FileSystemEventHandler):
    extension = '.sass'

    def __init__(self, outfiles, dirs, files=None, lib_dirs = None):
        self.outfiles = outfiles
        self.dirs = dirs
        self.lib_dirs = lib_dirs if lib_dirs else []
        self.source_files = get_source_files(self.dirs,extension=self.extension)
        if files:
            self.source_files += files
        super(FileSystemEventHandler, self).__init__()

    def write_out(self):
        try:
            lib_dirs = get_source_dirs(self.dirs)
            if self.lib_dirs:
                lib_dirs += self.lib_dirs
            lib_dirs = distinct(lib_dirs)

            compiler = SassCompiler(lib_dirs=lib_dirs)
            for sf in self.source_files:
                print "Reading %s" % os.path.abspath(sf)
                compiler.read_file(sf)


            normal_output = compiler.get_css()
            if 'normal' in self.outfiles:
                normal_out_filename = os.path.abspath(self.outfiles['normal'])
                print "Writing %s" % normal_out_filename
                with open(normal_out_filename, 'w') as of:
                    of.write(normal_output)

            if 'minified' in self.outfiles:
                minified_out_filename = os.path.abspath(self.outfiles['minified'])
                print "Writing %s" % minified_out_filename
                with open(minified_out_filename, 'w') as of:
                    of.write(minify(normal_output))

        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)

    def on_any_event(self, event):
        paths = []
        if hasattr(event, 'src_path'):
            paths += split_paths(event.src_path)
        if hasattr(event, 'dest_path'):
            paths += split_paths(event.dest_path)

        changed = False
        for p in paths:
            if p in self.source_files:
                changed = True
                break

            if os.path.abspath(os.path.dirname(p)) in self.lib_dirs and p.endswith(self.extension):
                changed = True
                break

        if changed:
            # One or more sass file changes, trying to regenerate output
            self.write_out()


class SassObserver(Observer):

    def add_output(self, outfiles, dirs=None, files=None, lib_dirs=None):

        if isinstance(outfiles, basestring):
            outfiles = {'normal': outfiles}

        if dirs is None:
            dirs = []

        if files is None:
            files = []

        if lib_dirs is None:
            lib_dirs = []

        dirs = distinct(split_paths(dirs))

        handler = IsassEventHandler(outfiles, dirs, files=files, lib_dirs=lib_dirs)
        watch_dirs = list(dirs)

        if lib_dirs:
            watch_dirs += split_paths(lib_dirs)
        if files:
            watch_dirs += split_paths([os.path.dirname(p) for p in files])

        watch_dirs = [wd for wd in distinct(watch_dirs) if os.path.exists(wd)]

        handler.write_out()
        for d in watch_dirs:
            self.schedule(handler, d, recursive=False)

    def add_manifest(self, manifest):
        manifest = Manifest(manifest)
        for taskname in manifest.get_task_names():
            task = manifest[taskname]
            dirs = []
            files = []
            for s in task.sources:
                if os.path.isfile(s):
                    files.append(s)
                else:
                    dirs.append(s)
            self.add_output(outfiles=task.output,
                            dirs=dirs,
                            files=files,
                            lib_dirs=task.libdirs)





__version__ = '0.1.10'

from .compiler import SassCompiler
from .watch import SassObserver
from .manifest import Manifest

def get_css(sass, lib_dirs=None):
    c = SassCompiler(lib_dirs=lib_dirs)
    c.read_string(sass)
    return c.get_css()

def get_scss(sass, lib_dirs=None):
    c = SassCompiler(lib_dirs=lib_dirs)
    c.read_string(sass)
    return c.get_scss()



# live(outputs,lib_dirs=None)

# TODO: scripts watchdog


__all__ = ['__version__',
           'get_css',
           'get_scss',
           'SassCompiler',
           'SassObserver',
           'Manifest']





# -*- coding: utf-8 -*-
'''
Created on:    Apr 1, 2014
@author:        vahid
'''
from pymlconf import ConfigManager, ConfigList, ConfigDict
from isass.helpers import get_source_dirs, get_source_files, distinct, minify
from isass import SassCompiler
import os.path



class Manifest(ConfigManager):
    def __init__(self,manifest_file):
        ConfigManager.__init__(self, files=manifest_file)
        manifest_dir = os.path.dirname(manifest_file)
        for key in self.get_task_names():
            manifest_config = self[key]
            manifest_config.sources = [i if i.startswith('/') else os.path.abspath(os.path.join(manifest_dir, i)) for i in manifest_config.sources]

            if hasattr(manifest_config,'libdirs'):
                manifest_config.libdirs = [i if i.startswith('/') else os.path.abspath(os.path.join(manifest_dir, i)) for i in manifest_config.libdirs]
            else:
                manifest_config.libdirs = ConfigList()

            if isinstance(manifest_config.output, basestring):
                manifest_config.output = ConfigDict({'normal': manifest_config.output})

            for k, v in manifest_config.output.items():
                if not v.startswith('/'):
                    manifest_config.output[k] = os.path.abspath(os.path.join(manifest_dir, v))

    def get_task_names(self):
        for k in self.keys():
            if isinstance(k,basestring) and not k.startswith('_'):
                yield k

    def process_task(self, taskname, extension=".sass"):
        task = self[taskname]
        dirs = []
        files = []
        for s in task.sources:
            if os.path.isfile(s):
                files.append(s)
            else:
                dirs.append(s)
        lib_dirs = get_source_dirs(dirs)
        if lib_dirs:
            lib_dirs += lib_dirs
        lib_dirs = distinct(lib_dirs)

        source_files = get_source_files(dirs,extension=extension)
        if files:
            source_files += files

        compiler = SassCompiler(lib_dirs=lib_dirs)
        for sf in source_files:
            #print "Reading %s" % os.path.abspath(sf)
            compiler.read_file(sf)


        normal_output = compiler.get_css()

        if isinstance(task.output, basestring):
            task.output = {'normal': task.output}

        if 'normal' in task.output:
            normal_out_filename = os.path.abspath(task.output['normal'])
            print "Writing %s" % normal_out_filename
            with open(normal_out_filename, 'w') as of:
                of.write(normal_output)

        if 'minified' in task.output:
            minified_out_filename = os.path.abspath(task.output['minified'])
            print "Writing %s" % minified_out_filename
            with open(minified_out_filename, 'w') as of:
                of.write(minify(normal_output))

    def write_outputs(self):
        for taskname in self.get_task_names():
            self.process_task(taskname)



# -*- coding: utf-8 -*-
'''
Created on:    Nov 10, 2013
@author:        vahid
'''
import os
import glob
from csscompressor import compress

def purifylines(text):
    return [ j for j in 
            [i.strip().replace('}','').replace('{','').replace('; ',';') for i in splitlines(text)] if j.strip()]

def splitlines(text):
    return [i.strip() for i in text.splitlines() if i.strip()]


def split_paths(p):
    if isinstance(p, basestring):
        res = p.split(',')
    else:
        res = p
    return [os.path.abspath(i.strip()) for i in res]

def distinct(l):
    if not l or not len(l):
        return []
    return list(set(l))

def get_source_files(sources,extension='.sass'):
    source_files = []
    for s in sources:
        p = os.path.abspath(s)
        if os.path.isfile(p):
            source_files.append(p)
        elif os.path.isdir(p):
            for _f in glob.iglob(os.path.join(p,'*%s' % extension)):
                source_files.append(os.path.abspath(_f))
    return sorted(distinct(source_files))

def get_source_dirs(sources):
    source_dirs = []
    for s in sources:
        if os.path.isdir(s):
            source_dirs.append(s)
        else:
            p = os.path.abspath(os.path.dirname(s))
            if os.path.isdir(p):
                source_dirs.append(p)
            
    return sorted(distinct(source_dirs))

def minify(css):
    return compress(css)



# -*- coding: utf-8 -*-
'''
Created on:    Nov 1, 2013
@author:        vahid
'''

import os
import StringIO
from isass.helpers import distinct


# attempt to load PySCSS module
try:
    import scss
    if 'Scss' in dir(scss):
        compile_scss = scss.Scss(scss_opts={'compress': True}).compile
except:
    compile_scss = None

class SassCompiler(object):
    def __init__(self,lib_dirs=None):
        self.lib_dirs = sorted(distinct(lib_dirs))
        self._scss_buffer = ''

    def scss_from_sass(self,sass,source_fname=None):
        output_buffer = StringIO.StringIO()
    
        state = {
          'prev_line': '',
          'line_buffer': '',
          'first_indent': None,
          'prev_indents': [],
        }
      
        # create a separate funtion for parsing the line so that we can call it again after the loop terminates
        def parse_line(line, i_line, state):
            # remove EOL character
            line = line.rstrip() 
          
            # check for line continuation
            if state['line_buffer']:
                line = state['line_buffer'] + line.lstrip()
            if line and line[-1] == ',':
                # line continues, skip this line and proceed to the next
                state['line_buffer'] = line + ' '
                return
            else:
                state['line_buffer'] = ''

            if state['prev_line'].strip().startswith('//'):
                state['prev_line']= state['prev_line'].replace('//', '/*')
            is_comment = state['prev_line'].strip().startswith('/*')
            if is_comment:
                text = state['prev_line'].rstrip()
                if not text.endswith('*/'):
                    state['prev_line'] = text + ' */'
        
            # mark indents from first non-whitespace column
            raw_indent = len(line) - len(line.lstrip())
            if state['first_indent'] is None:
                state['first_indent'] = raw_indent
            indent = raw_indent - state['first_indent']
        
            if indent == sum(state['prev_indents']):
        
                if '@import' in state['prev_line']:
                    import_fname = state['prev_line'].split()[1]
                    if import_fname.startswith('"'):
                        import_fname = import_fname[1:-1]

                    if not import_fname.endswith('.sass'):
                        import_fname += '.sass'

                    if self.lib_dirs:
                        for d in self.lib_dirs:
                            if os.path.isfile(os.path.join(d, import_fname)):
                                import_fname = os.path.join(d, import_fname)
                                break                

                    if not os.path.isfile(import_fname):
                        import_fname = os.path.join(os.path.dirname(source_fname), import_fname)

                        if not os.path.isfile(import_fname):
                            raise IOError('Error: @import {0} not found at {1}:{2}'.format(import_fname, source_fname, i_line))

                    state['prev_line'] = self.scss_from_sassfile(import_fname)
                
                elif not is_comment and state['prev_line']:
                    state['prev_line'] += ';'
        
            elif indent > sum(state['prev_indents']):
                # indentation is greater than previous: entered a new block
                state['prev_line'] += ' {'
                block_diff = indent - sum(state['prev_indents'])
                state['prev_indents'].append(block_diff)
            
            else:
                # indentation is shorter than previous: exited out of block
                if not is_comment and state['prev_line']:
                    state['prev_line'] += ';'
                # pull off prev_indents one-by-one and add }
                while len(state['prev_indents']) and indent < sum(state['prev_indents']):
                    state['prev_indents'].pop()
                    if sum(state['prev_indents']) < indent:
                        raise ValueError('Error: indentation mismatch at {0}:{1}'.format(source_fname, i_line+1))
                    state['prev_line'] += ' }' 
        
            if state['prev_line']:
                i = state['first_indent']
                left_padding = state['prev_line'][:i]
                if left_padding.strip():
                    raise ValueError('Error: indentation mismatch at {0}:{1}'.format(source_fname, i_line))
                output_buffer.write(state['prev_line'][i:] + '\n')
            
            state['prev_line'] = line
        
        i_line = 0
        for i_line, input_line in enumerate(sass.splitlines()):
            if input_line.strip():
                parse_line(input_line, i_line, state)
            
        # need this last pass to flush the buffered 'prev_line'
        parse_line('\n', i_line+1, state)
        
        return output_buffer.getvalue()

    def scss_from_sassfile(self,sass_fname):
        with open(sass_fname) as f:
            sass_text = f.read()
        return self.scss_from_sass(sass_text, source_fname=sass_fname)

    def read_string(self,sass):
        self._scss_buffer += self.scss_from_sass(sass)
    
    def read_file(self,fname):
        self._scss_buffer += self.scss_from_sassfile(fname)
        
    def get_scss(self):
        return self._scss_buffer
    
    def get_css(self):
        if compile_scss is None:
            raise "Error: couldn't load a SCSS compiler"
        return compile_scss(self._scss_buffer)




# -*- coding: utf-8 -*-
__author__ = 'vahid'
import argparse
import os
import sys
import time
from isass import helpers, SassObserver, SassCompiler, Manifest


parser = argparse.ArgumentParser(description='isass compiles SASS-indented-syntax into CSS or SCSS.')
parser.add_argument('sources', nargs='*', default=[], help='Source files or directories to process. default: standard input. example: `./*.sass` or `.`')
parser.add_argument('-o', '--output', metavar='OUTPUT' ,dest='output', help='Output file. default: standard output')
parser.add_argument('-c','--scss', dest='scss', action='store_true', help='Skip scss compilation, just return scss contents.')
parser.add_argument('-l','--lib-dir', dest='lib_dirs', nargs='?', action='append', help='Library dir to search for @imports.')
parser.add_argument('-e','--extension', dest='extension', nargs='?', default='*.sass',help='Search for this file extension.')
parser.add_argument('-m','--manifest-file', dest='manifest_file', nargs='+', help='Isass manifest file')
parser.add_argument('-w','--watch', dest='watch', action='store_true', help='Watch for modifications, and update output.')



def read_source():
    inp = ''
    if args.sources:
        source_files = helpers.get_source_files(args.sources, args.extension)
        for f in source_files:
            with open(f) as reader:
                inp += reader.read()
                inp += '\n'
    else:
        # reading from standard input
        inp += sys.stdin.read()

    return inp

def build(sass):
    compiler = SassCompiler(lib_dirs=args.lib_dirs)
    compiler.read_string(sass)
    if args.scss:
        return compiler.get_scss()
    else:
        return compiler.get_css()

def start_watchdog():

    # Asserts
    assert args.output, "Please provide an output file via -o option, Standard output was not supported in watch mode."
    source_msg = "Please provide one or more directory as sources, files and standard input was not supported in watch mode."
    assert len(args.sources) > 0, source_msg
    for s in args.sources:
        assert os.path.isdir(s), source_msg

    observer = SassObserver()
    observer.add_output(args.output, dirs=helpers.get_source_dirs(args.sources), lib_dirs=args.lib_dirs)
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def main():
    global args
    args = parser.parse_args()


    if args.watch:
        start_watchdog()
    else:
        #TODO: standard output
        if args.output:
            with open(args.output,'w') as outfile:
                outfile.write(build(read_source()))
        elif args.manifest_file:
            if isinstance(args.manifest_file, basestring):
                args.manifest_file = [args.manifest_file]
            for f in args.manifest_file:
                m = Manifest(f)
                m.write_outputs()
        else:
            print build(read_source())

